import datetime
from typing import List, Dict
from stixorm.module.authorise import authorised_mappings
import copy

import logging

from stixorm.module.typedb_lib.factories.auth_factory import get_auth_factory_instance
from stixorm.module.typedb_lib.factories.definition_factory import get_definition_factory_instance
from stixorm.module.typedb_lib.factories.import_type_factory import ImportType
from stixorm.module.typedb_lib.model.definitions import DefinitionName

stix_models = get_definition_factory_instance().lookup_definition(DefinitionName.STIX_21)
logger = logging.getLogger(__name__)


# ---------------------------------------------------
# 1.5) Sub Object Methods for adding common standard properties
#                -  e.g. stix-type, stix-id, name, description etc.
# --------------------------------------------------


def clean_props(total_props):
    """
        Future function to clean the list of properties
    Args:
        total_props []:

    Returns:
        total_props []:
    """
    # remove the properties that are mistakes

    return total_props


def add_property_to_typeql(prop, obj_tql, obj, prop_var_list):
    """
        Add a property by typeql
    Args:
        prop (): property
        obj_tql (): the tql that applies to this
        obj (): the stix object this is part of
        prop_var_list ():

    Returns:
        type_ql, the basic typeql statement for the property
        type_ql_props, the typeql statement that defines the property
        prop_var_list, a list of dicts describing the property
    """
    type_ql = type_ql_props = ''
    tql_prop_name = obj_tql[prop]
    # if property is defanged, summary or revoked, and the value is false, then don't add it to typedb_lib description
    if prop == "defanged" and obj.defanged == False:
        return type_ql, type_ql_props, prop_var_list
    elif prop == "revoked" and obj.revoked == False:
        return type_ql, type_ql_props, prop_var_list
    elif prop == "summary" and obj.summary == False:
        return type_ql, type_ql_props, prop_var_list

    # or else add the property to the typeql statement
    if isinstance(obj[prop], list):
        # if the property is a list, add each item to the typeql statement
        for i, instance in enumerate(obj[prop]):
            prop_var_dict = {}
            # import statements for each of the list items
            prop_var = '$' + prop + str(i)
            type_ql += ',\n has ' + tql_prop_name + ' ' + prop_var
            type_ql_props += '\n ' + prop_var + ' ' + val_tql(instance) + ';'
            prop_var_dict["prop_var"] = prop_var
            prop_var_dict["prop"] = prop
            prop_var_dict["index"] = i
            prop_var_list.append(prop_var_dict)
    else:
        prop_var_dict = {}
        # import statements for a single value
        prop_var = '$' + tql_prop_name
        type_ql += ',\n has ' + tql_prop_name + ' ' + prop_var
        type_ql_props += '\n ' + prop_var + ' ' + val_tql(obj[prop]) + ';'
        prop_var_dict["prop_var"] = prop_var
        prop_var_dict["prop"] = prop
        prop_var_dict["index"] = -1
        prop_var_list.append(prop_var_dict)

    return type_ql, type_ql_props, prop_var_list


# ---------------------------------------------------
# 1.6) Sub Object Methods for adding embedded structures
#                -  e.g. hasehs, kill-chain-phases, created_by, external_references, object_marking_refs etc.
# --------------------------------------------------
# Giant Switch statement to add the embedded relations to the typeql statement

def add_relation_to_typeql(rel,
                           obj,
                           obj_var,
                           prop_var_list,
                           import_type: ImportType,
                           inc,
                           protocol: str):
    """
        Top level function to add one of the sub objects to the stix object
    Args:
        rel (): the relation object to add
        obj (): the stix object to add it too
        obj_var (): the typeql variable string
        prop_var_list (): the property variable list
        import_type: the dict describing import preferences
        inc (): an incrementing variable that is used to add to the var string

    Returns:
        match: the typeql match strings
        insert: the typeql insert string
    """
    logger.debug(f'===============\n=====================\n===================\n')
    logger.debug(f'rel {rel}')
    logger.debug(f'obj[rel] {obj[rel]}')
    logger.debug(f'obj {obj}')
    logger.debug(f'obj_Var {obj_var}')
    logger.debug(f'\nprop var list {prop_var_list}')
    auth_factory = get_auth_factory_instance()
    auth = auth_factory.get_auth_for_import(import_type)
    dep_list = []
    logger.debug("\nstarting into choices")
    if rel == "granular_markings":
        logger.debug("in granular")
        match, insert = granular_markings(rel, obj[rel], obj_var, prop_var_list)

    # hashes type
    elif (rel == "hashes"
          or rel == "file_header_hashes"):
        logger.debug("in hashes")
        match, insert = hashes(rel, obj[rel], obj_var)

    # insert key value store
    elif rel in auth["reln_name"]["key_value_relations"]:
        logger.debug("in key value")
        match, insert = key_value_store(rel, obj[rel], obj_var, import_type)

    # insert list of object relation
    elif rel in auth["reln_name"]["list_of_objects"]:
        logger.debug("list of objects")
        match, insert, dep_list = list_of_object(rel, obj[rel], obj_var, inc, import_type, protocol)

    # insert embedded relations based on stix-id
    elif rel in auth["reln_name"]["embedded_relations"]:
        logger.debug("embedded")
        match, insert, dep_list = embedded_relation(rel, obj[rel], obj_var, inc, import_type, protocol)

    # insert plain sub-object with relation
    elif (rel == "x509_v3_extensions"
          or rel == "optional_header"):
        logger.debug("X509")
        match, insert, dep_list = load_object(rel, obj[rel], obj_var, inc, import_type, protocol)

    # insert  SCO Extensions here, a possible dict of sub-objects
    elif rel in auth["reln_name"]["extension_relations"] or rel == "extensions":
        logger.debug("extension")
        match, insert, dep_list = extensions(rel, obj[rel], obj_var, inc, import_type, protocol)

    # ignore the following relations as they are already processed, for Relationships, Sightings and Extensions
    elif rel in auth["reln_name"]["standard_relations"] or rel == "definition" or "definition_type":
        logger.debug("standard")
        match = insert = ''

    else:
        logger.debug(f'relation type not known, ignore if "source_ref" or "target_ref" -> {rel}')
        logger.debug("in else")
        match = insert = ""

    return match, insert, dep_list


# ---------------------------------------------------
# Methods for adding the embedded structures to the typeql statement
# --------------------------------------------------
# generic methods


def extensions(prop_name: str,
               prop_dict,
               parent_var,
               inc,
               import_type: ImportType,
               protocol: str):
    """
        Create the Typeql for the extensions sub object
    Args:
        prop_name (): the name of the extension
        prop_dict (): the dict for the extension
        parent_var (): the var of the Stix object that is the owner
        import_type: the dict describing import preferences

    Returns:
        match: the typeql match string
        insert: the typeql insert string
    """
    auth_factory = get_auth_factory_instance()
    auth = auth_factory.get_auth_for_import(import_type)
    match = insert = ''
    dep_list = []
    # for each key in the dict (extension type)
    # logger.debug('--------------------- extensions ----------------------------')
    for ext_type in prop_dict:
        for ext_type_ql in auth["reln"]["extension_relations"]:
            if ext_type == ext_type_ql["stix"]:
                match2, insert2, dep_list2 = load_object(ext_type, prop_dict[ext_type], parent_var, inc, import_type, protocol)
                match = match + match2
                insert = insert + insert2
                dep_list = dep_list + dep_list2
                break

    return match, insert, dep_list


def load_object(prop_name: str,
                prop_dict,
                parent_var: str,
                inc,
                import_type: ImportType,
                protocol: str):
    """
        Create the Typeql for a sub object
    Args:
        prop_name (): the name of the extension
        prop_dict (): the dict for the extension
        parent_var (): the var of the Stix object that is the owner
        import_type: the dict describing import preferences

    Returns:
        match: the typeql match string
        insert: the typeql insert string
    """
    auth_factory = get_auth_factory_instance()
    auth = auth_factory.get_auth_for_import(import_type)
    match = insert = type_ql = type_ql_props = ''
    # as long as it is predefined, history the object
    # logger.debug('------------------- history object ------------------------------')
    for prop_type in auth["reln"]["extension_relations"]:
        if prop_name == prop_type["stix"]:
            tot_prop_list = [tot for tot in prop_dict.keys()]
            obj_type = prop_type["object"]
            obj_tql = copy.deepcopy(auth["sub_objects"][obj_type])
            obj_var = '$' + obj_type
            reln = prop_type["relation"]
            rel_var = '$' + reln
            rel_owner = prop_type["owner"]
            rel_pointed_to = prop_type["pointed-to"]
            type_ql += ' ' + obj_var + ' isa ' + obj_type
            # Split them into properties and relations
            total_props = prop_dict._inner
            logger.debug(f'load object properties: {total_props}')
            properties, relations = split_on_activity_type(total_props, obj_tql)
            prop_var_list = []
            dep_list = []
            logger.debug(f'load object relations: {relations}')
            for prop in properties:
                # split off for properties processing
                type_ql2, type_ql_props2, prop_var_list = add_property_to_typeql(prop, obj_tql, prop_dict,
                                                                                 prop_var_list)
                # then add them all together
                type_ql += type_ql2
                type_ql_props += type_ql_props2
                # add a terminator on the end of the insert statement
            type_ql += ";\n" + type_ql_props + "\n\n"

            # add each of the relations to the match and insert statements
            logger.debug(f'load object relations: {relations}')
            for rel in relations:
                # split off for relation processing
                logger.debug(f'load object relation: {rel}, protocol: {protocol}')
                match2, insert2, dep_list2 = add_relation_to_typeql(rel, prop_dict, obj_var, prop_var_list, import_type, inc, protocol)
                # then add it back together    
                match = match + match2
                insert = insert + "\n" + insert2
                dep_list = dep_list + dep_list2

            # finally, connect the local object to the parent object
            type_ql += ' ' + rel_var + ' (' + rel_owner + ':' + parent_var
            type_ql += ', ' + rel_pointed_to + ':' + obj_var + ')'
            type_ql += ' isa ' + reln + ';\n'
            break

    insert = type_ql + "\n" + insert
    return match, insert, dep_list


def list_of_object(prop_name: str,
                   prop_value_list: List[str],
                   parent_var,
                   inc,
                   import_type: ImportType,
                   protocol: str):
    """
        Create the Typeql for the list of object sub object
    Args:
        prop_name (): the name of the object
        prop_value_list (): the list of object
        parent_var (): the var of the Stix object that is the owner
        import_type: the dict describing import preferences

    Returns:
        match: the typeql match string
        insert: the typeql insert string
    """
    auth_factory = get_auth_factory_instance()
    auth = auth_factory.get_auth_for_import(import_type)
    for config in auth["reln"]["list_of_objects"]:
        if config["name"] == prop_name:
            rel_typeql = config["typeql"]
            role_owner = config["owner"]
            role_pointed = config["pointed_to"]
            typeql_obj = config["object"]
            break

    if typeql_obj in auth["sub_objects"]:
        obj_props_tql = copy.deepcopy(auth["sub_objects"][typeql_obj])
    else:
        raise ValueError("no sub-object available")

    lod_list = []
    dep_list = []
    match = rel_insert = rel_match = insert = ''
    for i, dict_instance in enumerate(prop_value_list):
        lod_var = '$' + typeql_obj + str(i)
        lod_list.append(lod_var)
        insert += lod_var + ' isa ' + typeql_obj
        for key in dict_instance:
            typeql_prop = obj_props_tql[key]
            if typeql_prop == '':
                rel_match2, rel_insert2, dep_list2 = add_relation_to_typeql(key, dict_instance, lod_var, [], import_type, i, protocol)
                rel_insert += rel_insert2
                rel_match += rel_match2
                dep_list = dep_list + dep_list2
            else:
                insert += ',\n has ' + typeql_prop + ' ' + val_tql(dict_instance[key])
        insert += ';\n'

    insert += '\n $' + rel_typeql + ' (' + role_owner + ':' + parent_var
    for lod_var in lod_list:
        insert += ', ' + role_pointed + ':' + lod_var

    insert += ') isa ' + rel_typeql + ';\n' + rel_insert
    match += rel_match
    return match, insert, dep_list


def key_value_store(prop,
                    prop_value_dict,
                    obj_var,
                    import_type: ImportType):
    """
        Create the Typeql for the key-value store sub object
    Args:
        prop (): the name of the object
        prop_value_dict (): the dict of object
        obj_var (): the var of the Stix object that is the owner
        import_type: the dict describing import preferences

    Returns:
        match: the typeql match string
        insert: the typeql insert string
    """
    auth_factory = get_auth_factory_instance()
    auth = auth_factory.get_auth_for_import(import_type)
    for config in auth["reln"]["key_value_relations"]:
        if config["name"] == prop:
            rel_typeql = config["typeql"]
            role_owner = config["owner"]
            role_pointed = config["pointed_to"]
            d_key = config["key"]
            d_value = config["value"]
            break

    match = ''
    insert = '\n'
    field_var_list = []
    for i, key in enumerate(prop_value_dict):
        a_value = prop_value_dict[key]
        key_var = ' $' + d_key + str(i)
        field_var_list.append(key_var)
        insert += key_var + ' isa ' + d_key + '; ' + key_var + ' "' + key + '";\n'
        if isinstance(a_value, list):
            for j, n in enumerate(a_value):
                value_var = ' $' + d_value + str(j)
                insert += key_var + ' ' + 'has ' + d_value + ' "' + str(n) + '";\n'
        else:
            value_var = ' $' + d_value + str(i)
            insert += key_var + ' ' + 'has ' + d_value + ' "' + str(a_value) + '";\n'

    insert += ' $' + rel_typeql + ' (' + role_owner + ':' + obj_var
    for var in field_var_list:
        insert += ', ' + role_pointed + ':' + var
    insert += ') isa ' + rel_typeql + ';\n\n'
    return match, insert


# specific methods
def hashes(prop_name, prop_dict, parent_var):
    """
        Create the Typeql for the hashes sub object
    Args:
        prop_name (): the name of the object
        prop_dict (): the dict of object
        parent_var (): the var of the Stix object that is the owner

    Returns:
        match: the typeql match string
        insert: the typeql insert string
    """
    match = insert = ''
    hash_var_list = []
    for i, key in enumerate(prop_dict):
        hash_var = '$hash' + str(i)
        hash_var_list.append(hash_var)
        if key in stix_models.get_sub_objects("hash_typeql_dict"):
            insert += ' ' + hash_var + ' isa ' + stix_models.get_sub_objects("hash_typeql_dict")[
                key] + ', has hash-value ' + val_tql(prop_dict[key]) + ';\n'
        else:
            logger.error(f'Unknown hash type {key}')

    # insert the hash objects into the hashes relation with the parent object
    insert += '\n $hash_rel (owner:' + parent_var
    for hash_var in hash_var_list:
        insert += ', pointed-to:' + hash_var

    insert += ') isa hashes;\n'
    return match, insert


def granular_markings(prop_name, prop_value_List, parent_var, prop_var_list):
    """
        Create the Typeql for the granular markings sub object
    Args:
        prop_name (): the name of the object
        prop_value_List (): the list of object values
        parent_var (): the var of the Stix object that is the owner
        prop_var_list: the list of property variables

    Returns:
        match: the typeql match string
        insert: the typeql insert string
    """
    match = insert = ''
    for i, prop_dict in enumerate(prop_value_List):
        # setup and match in the marking, based on its id
        m_id = prop_dict['marking_ref']
        m_var = '$marking' + str(i)
        g_var = '$granular' + str(i)
        match += ' ' + m_var + ' isa marking-definition, has stix-id ' + '"' + m_id + '";\n'
        insert += ' ' + g_var + ' (marking:' + m_var + ', object:' + parent_var
        prop_list = prop_dict['selectors']
        for selector in prop_list:
            selector_var = get_selector_var(selector, prop_var_list)
            insert += ', marked:' + selector_var

        insert += ') isa granular-marking;\n'

    return match, insert


def get_selector_var(selector, prop_var_list):
    """
        Get the typeql variable for the property
    Args:
        selector (): the property to select
        prop_var_list (): the variable list to select from

    Returns:
        selector_var: the typeql variable for the property
    """
    if selector[-1] == ']':
        text = selector.split(".")
        selector = text[0]
        index = int(text[1][1])
    else:
        selector = selector
        index = -1

    # logger.debug(f'selector after processing -> {selector}, index after procesing -> {index}')
    for prop_var_dict in prop_var_list:
        if selector == prop_var_dict['prop'] and index == prop_var_dict['index']:
            selector_var = prop_var_dict['prop_var']
            break

    return selector_var


# ---------------------------------------------------
#        EMBEDDED RELATION METHODS
# ---------------------------------------------------
# object_refs
# sample_refs
# sample_ref
# host_vm_ref
# operating_system_ref
# installed_software_refs
# analysis_sco_refs
# etc.

def  embedded_relation(prop,
                      prop_value,
                      obj_var,
                      inc: int,
                      import_type: ImportType,
                      protocol: str):
    """
        Create the Typeql for the embedded relation sub object
    Args:
        prop (): the name of the object
        prop_value (): the value of object
        obj_var (): the var of the Stix object that is the owner

    Returns:
        match: the typeql match string
        insert: the typeql insert string
    """
    logger.debug("I'm in embedded")
    auth_factory = get_auth_factory_instance()
    auth = auth_factory.get_auth_for_import(import_type)
    for ex in auth["reln"]["embedded_relations"]:
        if ex["rel"] == prop:
            owner = ex["owner"]
            pointed_to = ex["pointed-to"]
            relation = ex["typeql"]
            break

    prop_var_list = []
    dep_list = []
    match = ''
    if inc == -1:
        inc_add = ''
    else:
        inc_add = str(inc)
    # if the prop_value is a list, then match in each item
    logger.debug(f'\n1\n')
    if isinstance(prop_value, list):
        dep_list = prop_value
        logger.debug(f'deplist {dep_list}')
        for i, prop_v in enumerate(prop_value):
            prop_type = get_source_from_id(prop_v, protocol, import_type)
            if prop_type == 'relationship':
                prop_type = 'stix-core-relationship'
            prop_var = '$' + prop_type + str(i) + inc_add
            prop_var_list.append(prop_var)
            match += ' ' + prop_var + ' isa ' + prop_type + ', has stix-id ' + '"' + prop_v + '";\n'
    # else, match in the single prop_value
    else:
        dep_list.append(prop_value)
        prop_type = get_source_from_id(prop_value, protocol, import_type)
        logger.debug(f'deplist {dep_list}')
        if prop_type == 'relationship':
            prop_type = 'stix-core-relationship'
        prop_var = '$' + prop_type + inc_add
        prop_var_list.append(prop_var)
        match += ' ' + prop_var + ' isa ' + prop_type + ', has stix-id ' + '"' + prop_value + '";\n'

    # Then setup and insert the relation
    logger.debug(f'\n2\n')
    insert = '\n $' + relation + inc_add + ' (' + owner + ':' + obj_var
    for prop_var in prop_var_list:
        insert += ', ' + pointed_to + ':' + prop_var
    insert += ') isa ' + relation + ';\n'
    return match, insert, dep_list


def get_source_from_id(stid: str,
                       protocol: str,
                       import_type: ImportType):
    """
        Get the source of the stix object
    Args:
        stid (): the stix-id of the object

    Returns:
        source: the source of the object
    """
    tmp_source = stid.split('--')[0]
    auth_factory = get_auth_factory_instance()
    auth = auth_factory.get_auth_for_import(import_type)
    source = ""
    for model in auth["conv"]["sdo"]:
        if model["protocol"] == protocol and model["type"] == tmp_source:
            source = model["typeql"]
            return source
    for model in auth["conv"]["sro"]:
        if model["protocol"] == protocol and model["type"] == tmp_source:
            source = model["typeql"]
            if source == 'relationship' or source == "attack-relation":
                source = 'stix-core-relationship'
            return source
    for model in auth["conv"]["sco"]:
        if model["protocol"] == protocol and model["type"] == tmp_source:
            source = model["typeql"]
            return source
    for model in auth["conv"]["meta"]:
        if model["protocol"] == protocol and model["type"] == tmp_source:
            source = model["typeql"]
            return source
    for model in auth["conv"]["sdo"]:
        if model["type"] == tmp_source:
            source = model["typeql"]
            return source
    for model in auth["conv"]["sro"]:
        if model["type"] == tmp_source:
            source = model["typeql"]
            return source
    for model in auth["conv"]["sco"]:
        if model["type"] == tmp_source:
            source = model["typeql"]
            return source
    for model in auth["conv"]["meta"]:
        if model["type"] == tmp_source:
            source = model["typeql"]
            return source
    return source


def get_embedded_match(source_id: str,
                       i: int,
                       protocol: str,
                       import_type: ImportType):
    """
        Assemble the typeql variable and match statement given the stix-id, and the increment
    Args:
        source_id (): stix-id to use
        i (): number of times this type of object has been used

    Returns:
        source_var, the typeql string of the variable
        match, the typeql match statement
    """
    source_type = get_source_from_id(source_id, protocol, import_type)
    source_var = '$' + source_type + str(i)
    if source_type == 'relationship' or source_type == "attack-relation":
        source_type = 'stix-core-relationship'
    match = f' {source_var} isa {source_type}, has stix-id "{source_id}";\n'
    return source_var, match


def get_full_object_match(source_id: str, protocol: str, import_type: ImportType):
    """
        Return a typeql match statement for this stix object
    Args:
        source_id (): the stix-id to look for

    Returns:
        source_var, the typeql string of the variable
        match, the typeql match statement
    """
    source_var, match = get_embedded_match(source_id, 0, protocol, import_type)
    match += source_var + ' has $properties;\n'
    # match += '$embedded (owner:' + source_var + ', pointed-to:$point ) isa embedded;\n'
    return source_var, match


# ---------------------------------------------------
# 1.7) Helper Methods for 
#           - converting a Python value --> typeql string
#           - splitting a list of total properties into properties and relations
# ---------------------------------------------------


def val_tql(val):
    """
        Modify the value used in a typeql statement, depending on its type
    Args:
        val (): the value being used

    Returns:
        val: the value formatted for typeql
    """
    if isinstance(val, str):
        replaced_val = val.replace('"', "'")
        replaced_val2 = replaced_val.replace('\\', '\\\\')
        return '"' + replaced_val2 + '"'
    elif isinstance(val, bool):
        return str(val).lower()
    elif isinstance(val, int):
        return str(val)
    elif isinstance(val, float):
        return str(val)
    elif isinstance(val, datetime.datetime):
        return str(val.strftime("%Y-%m-%dT%H:%M:%S.%f")[:-3])
    else:
        return logger.error(f'value  not supported: {val}')


def split_on_activity_type(total_props: dict, obj_tql: Dict[str, str]) -> [List[str], List[str]]:
    """
        Split the Stix object properties into flat properties and sub objects
    Args:
        total_props (): the total properties for this object
        obj_tql (): the mapping dict for this object

    Returns:
        prop_list, a list of the flat properties
        rel_list, a list of the sub objects
    """
    prop_list = []
    rel_list = []
    logger.debug("@@@@@@@@@@@@@@@@@@@@@@ splitting @@@@@@@@@@@@@@@")
    logger.debug("========================================")
    logger.debug(f'total props: {total_props}')
    # for k, v in total_props.items():
    #     logger.debug(k, v)
    # logger.debug("=========================================")
    logger.debug("========================================")
    logger.debug(f'obj tql: {obj_tql}')
    # for k, v in obj_tql.items():
    #     logger.debug(k, v)
    # logger.debug("=========================================")
    logger.debug("@@@@@@@@@@@@@@@@@@@@@@ end splitting @@@@@@@@@@@@@@@")
    for prop in total_props:
        tql_prop_name = obj_tql[prop]
        logger.debug(f'prop {prop}, object tql -> {tql_prop_name}')

        if tql_prop_name == "":
            rel_list.append(prop)
            # logger.debug(f'Im a rel --> {prop},        tql --> {tql_prop_name}')
        else:
            prop_list.append(prop)
            # logger.debug(f'Im a prop --> {prop},        tql --> {tql_prop_name}')

    return prop_list, rel_list

###################################################################################################



import json
import os
import uuid
import copy
from enum import Enum
from typing import Optional
from stix2.properties import Property
from stix2.utils import STIXTypeClass
from stix2.exceptions import CustomContentError
from stix2.base import _STIXBase
from stix2.properties import (
    DictionaryProperty, ExtensionsProperty, IDProperty, IntegerProperty, ListProperty,
    OpenVocabProperty, ReferenceProperty, StringProperty,
    TimestampProperty, TypeProperty, EmbeddedObjectProperty, ObservableProperty
)


DEFAULT_VERSION = '2.1'
ERROR_INVALID_ID = (
    "not a valid STIX identifier, must match <object-type>--<UUID>: {}"
)

def get_ext_class(key, spec_version):
    defin = Definitions()
    list_of_ext = defin.sub_objects
    for ext in list_of_ext:
        if ext["type"] == key:
            pass
           # return os_threat_models[ext["class"]]



# TODO: Kestrel was missing from original definition, does this need to be fixed?
def get_libraries():
    return ["stix", "attack", "os-threat", "cacao"]


def get_type_from_id(stix_id):
    return stix_id.split('--', 1)[0]


def _check_uuid(uuid_str, spec_version):
    """
    Check whether the given UUID string is valid with respect to the given STIX
    spec version.  STIX 2.0 requires UUIDv4; 2.1 only requires the RFC 4122
    variant.
    :param uuid_str: A UUID as a string
    :param spec_version: The STIX spec version
    :return: True if the UUID is valid, False if not
    :raises ValueError: If uuid_str is malformed
    """
    uuid_obj = uuid.UUID(uuid_str)

    ok = uuid_obj.variant == uuid.RFC_4122
    if ok and spec_version == "2.0":
        ok = uuid_obj.version == 4

    return ok


def _validate_id(id_, spec_version, required_prefix):
    """
    Check the STIX identifier for correctness, raise an exception if there are
    errors.
    :param id_: The STIX identifier
    :param spec_version: The STIX specification version to use
    :param required_prefix: The required prefix on the identifier, if any.
        This function doesn't add a "--" suffix to the prefix, so callers must
        add it if it is important.  Pass None to skip the prefix check.
    :raises ValueError: If there are any errors with the identifier
    """
    if required_prefix:
        if not id_.startswith(required_prefix):
            raise ValueError("must start with '{}'.".format(required_prefix))

    try:
        if required_prefix:
            uuid_part = id_[len(required_prefix):]
        else:
            idx = id_.index("--")
            uuid_part = id_[idx+2:]

        result = _check_uuid(uuid_part, spec_version)
    except ValueError:
        # replace their ValueError with ours
        raise ValueError(ERROR_INVALID_ID.format(id_))

    if not result:
        raise ValueError(ERROR_INVALID_ID.format(id_))


def is_stix_type(obj_type, spec_version, valid_types):
    if obj_type in valid_types:
        return True
    else:
        return False


class ThreatReference(Property):

    _WHITELIST, _BLACKLIST = range(2)

    def __init__(self, valid_types=None, invalid_types=None, spec_version=DEFAULT_VERSION, **kwargs):
        """
        references sometimes must be to a specific object type
        """
        self.spec_version = spec_version

        if (valid_types is not None and invalid_types is not None) or \
                (valid_types is None and invalid_types is None):
            raise ValueError(
                "Exactly one of 'valid_types' and 'invalid_types' must be "
                "given",
            )

        if valid_types and not isinstance(valid_types, list):
            valid_types = [valid_types]
        elif invalid_types and not isinstance(invalid_types, list):
            invalid_types = [invalid_types]

        if valid_types is not None and len(valid_types) == 0:
            raise ValueError("Impossible type constraint: empty whitelist")

        # Divide type requirements into generic type classes and specific
        # types.  With respect to strings, values recognized as STIXTypeClass
        # enum names are generic; all else are specifics.
        self.valid_types = valid_types
        self.invalid_types = invalid_types

        super(ThreatReference, self).__init__(**kwargs)

    def clean(self, value, allow_custom):
        if isinstance(value, _STIXBase):
            value = value.id
        value = str(value)

        _validate_id(value, self.spec_version, None)

        obj_type = get_type_from_id(value)

        type_ok = is_stix_type(obj_type, self.spec_version, self.valid_types)

        if not type_ok:
            raise ValueError(
                "The type-specifying prefix '%s' for this property is %s"
                % obj_type,
            )

        return value, allow_custom



def _get_dict(data):
    """Return data as a dictionary.

    Input can be a dictionary, string, or file-like object.
    """

    if type(data) is dict:
        return data
    else:
        try:
            return json.loads(data)
        except TypeError:
            pass
        try:
            return json.load(data)
        except AttributeError:
            pass
        try:
            return dict(data)
        except (ValueError, TypeError):
            raise ValueError("Cannot convert '%s' to dictionary." % str(data))


class ThreatExtensionsProperty(DictionaryProperty):
    """Property for representing extensions on Observable objects.
    """

    def __init__(self, spec_version=DEFAULT_VERSION, required=False):
        super(ThreatExtensionsProperty, self).__init__(spec_version=spec_version, required=required)

    def clean(self, value, allow_custom):
        try:
            dictified = _get_dict(value)
            # get deep copy since we are going modify the dict and might
            # modify the original dict as _get_dict() does not return new
            # dict when passed a dict
            dictified = copy.deepcopy(dictified)
        except ValueError:
            raise ValueError("The extensions property must contain a dictionary")

        has_custom = False
        for key, subvalue in dictified.items():
            cls = get_ext_class(key, self.spec_version)
            if cls:
                if isinstance(subvalue, dict):
                    ext = cls(allow_custom=False, **subvalue)
                elif isinstance(subvalue, cls):
                    # If already an instance of the registered class, assume
                    # it's valid
                    ext = subvalue
                else:
                    raise TypeError(
                        "Can't create extension '{}' from {}.".format(
                            key, type(subvalue),
                        ),
                    )

                has_custom = has_custom or ext.has_custom

                if not allow_custom and has_custom:
                    raise CustomContentError(
                        "custom content found in {} extension".format(
                            key,
                        ),
                    )

                dictified[key] = ext

            else:
                # If an unregistered "extension-definition--" style extension,
                # we don't know what's supposed to be in it, so we can't
                # determine whether there's anything custom.  So, assume there
                # are no customizations.  If it's a different type of extension,
                # non-registration implies customization (since all spec-defined
                # extensions should be pre-registered with the library).

                if key.startswith('extension-definition--'):
                    _validate_id(
                        key, self.spec_version, 'extension-definition--',
                    )
                elif allow_custom:
                    has_custom = True
                else:
                    raise CustomContentError("Can't parse unknown extension type: {}".format(key))

                dictified[key] = subvalue

        return dictified, has_custom



from stixorm.module.typedb_lib.factories.definition_factory import get_definition_factory_instance
from stixorm.module.typedb_lib.model.definitions import DefinitionName

stix_models = get_definition_factory_instance().lookup_definition(DefinitionName.STIX_21)
attack_models = get_definition_factory_instance().lookup_definition(DefinitionName.ATTACK)
os_threat_models = get_definition_factory_instance().lookup_definition(DefinitionName.OS_THREAT)
cacao_models = get_definition_factory_instance().lookup_definition(DefinitionName.CACAO)
kestrel_models = get_definition_factory_instance().lookup_definition(DefinitionName.KESTREL)
us_dod_models = get_definition_factory_instance().lookup_definition(DefinitionName.US_DoD)

class AuthTypeFactory:


    def __init__(self):
        self.auth_types = {}
        self.auth_types["stix"] = stix_models
        self.auth_types["attack"] = attack_models
        self.auth_types["os_threat"] = os_threat_models
        self.auth_types["cacao"] = cacao_models
        self.auth_types["kestrel"] = kestrel_models
        self.auth_types["us_dod"] = us_dod_models





import re

ts1 = "$created-by0 (created:$threat-actor, creator:$identity0) isa created-by;\n \n $object-marking1 (marked:$threat-actor, marking:$marking-definition01) isa object-marking;"

ts2 = '$threat-actor isa threat-actor, has stix-id "threat-actor--8b6297fe-cae7-47c6-9256-5584b417849c";'

def get_local_relns(dep_insert):
    m = re.findall(r"(\$[a-z\-0-9]+)\s+(.+)?isa\s+", dep_insert,re.MULTILINE)
    #logger.info(f'local 1 -> {m}')
    if m:
        #logger.info(f'local -> {m}')
        return [g[0] for g in m]
    else:
        return []


def get_object_variable(core_ql):
    m = re.findall(r"[a-z\-\_]+;$",core_ql,re.MULTILINE)
    return m

# if this file is run directly, then start here
if __name__ == '__main__':
    logger.info("ts1=================")
    get_local_relns(ts2)
    logger.info("ts2====================")
    get_object_variable(ts1)


#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Loader module for all STIX data model
"""

__author__ = "Paolo Di Prodi"
__credits__ = ["Brett Forbes"]
__license__ = "GPL"
__version__ = "0.1.0"
__maintainer__ = "Paolo Di Prodi"
__email__ = "paolo@priam.ai"
__status__ = "Production"

import json
from glob import glob

import pathlib
from loguru import logger
import os
from pathlib import Path
from stixorm.module.definitions.stix21.classes import (
    Note, ObservedData, Incident, Report
)
from stix2.v21.sdo import (
    AttackPattern, Campaign, CourseOfAction, CustomObject, Grouping, Identity,
    Indicator, Infrastructure, IntrusionSet, Location, Malware,
    MalwareAnalysis, Opinion, ThreatActor, Tool,
    Vulnerability,
)
from stix2.v21.observables import (
    URL, AlternateDataStream, ArchiveExt, Artifact, AutonomousSystem,
    CustomObservable, Directory, DomainName, EmailAddress, EmailMessage,
    EmailMIMEComponent, File, HTTPRequestExt, ICMPExt, IPv4Address,
    IPv6Address, MACAddress, Mutex, NetworkTraffic, NTFSExt, PDFExt, Process,
    RasterImageExt, SocketExt, Software, TCPExt, UNIXAccountExt, UserAccount,
    WindowsPEBinaryExt, WindowsPEOptionalHeaderType, WindowsPESection,
    WindowsProcessExt, WindowsRegistryKey, WindowsRegistryValueType,
    WindowsServiceExt, X509Certificate, X509V3ExtensionsType,
)
from stix2.v21.sro import Relationship, Sighting
from stix2.v21.common import MarkingDefinition

name = "stix21"
class_model = {}
class_model["sro"] = {
    "Relationship": Relationship,
    "Sighting": Sighting
}
class_model["sdo"] = {
    "AttackPattern":AttackPattern,
    "Campaign":Campaign,
    "CourseOfAction":CourseOfAction,
    "CustomObject":CustomObject,
    "Grouping":Grouping,
    "Identity":Identity,
    "Incident":Incident,
    "Indicator":Indicator,
    "Infrastructure":Infrastructure,
    "IntrusionSet":IntrusionSet,
    "Location":Location,
    "Malware":Malware,
    "MalwareAnalysis":MalwareAnalysis,
    "Note":Note,
    "ObservedData":ObservedData,
    "Opinion":Opinion,
    "Report":Report,
    "ThreatActor":ThreatActor,
    "Tool":Tool,
    "Vulnerability":Vulnerability,
}
class_model["sub"] = {
    "AlternateDataStream":AlternateDataStream,
    "ArchiveExt":ArchiveExt,
    "EmailMIMEComponent": EmailMIMEComponent,
    "HTTPRequestExt":HTTPRequestExt,
    "ICMPExt":ICMPExt,
    "NTFSExt":NTFSExt,
    "PDFExt":PDFExt,
    "RasterImageExt":RasterImageExt,
    "SocketExt":SocketExt,
    "TCPExt":TCPExt,
    "UNIXAccountExt":UNIXAccountExt,
    "WindowsPEBinaryExt":WindowsPEBinaryExt,
    "WindowsPEOptionalHeaderType":WindowsPEOptionalHeaderType,
    "WindowsPESection":WindowsPESection,
    "WindowsProcessExt":WindowsProcessExt,
    "WindowsRegistryValueType":WindowsRegistryValueType,
    "WindowsServiceExt":WindowsServiceExt,
    "X509V3ExtensionsType":X509V3ExtensionsType
}
class_model["sco"] = {
    "URL":URL,
    "Artifact":Artifact,
    "AutonomousSystem":AutonomousSystem,
    "CustomObservable":CustomObservable,
    "Directory":Directory,
    "DomainName":DomainName,
    "EmailAddress":EmailAddress,
    "EmailMessage":EmailMessage,
    "File":File,
    "IPv4Address":IPv4Address,
    "IPv6Address":IPv6Address,
    "MACAddress":MACAddress,
    "Mutex":Mutex,
    "NetworkTraffic":NetworkTraffic,
    "Process":Process,
    "Software":Software,
    "UserAccount":UserAccount,
    "WindowsRegistryKey":WindowsRegistryKey,
    "X509Certificate":X509Certificate,
}
class_model["meta"] = {
    "MarkingDefinition":MarkingDefinition
}




import logging
import traceback
from typing import List, Iterator
from typedb.api.answer.concept_map import ConceptMap
from typedb.api.connection.client import TypeDBClient
from typedb.api.connection.session import SessionType, TypeDBSession
from typedb.api.connection.transaction import TransactionType, TypeDBTransaction
from typedb.api.query.future import QueryFuture
from typedb.api.query.query_manager import QueryManager
from typedb.client import TypeDB
from typedb.stream.bidirectional_stream import BidirectionalStream

from stixorm.module.typedb_lib.logging import log_delete_layer, log_add_layer
from stixorm.module.typedb_lib.instructions import Instructions

logger = logging.getLogger(__name__)



def build_insert_query(layer):
    dep_match = layer["dep_match"]
    dep_insert = layer["dep_insert"]
    indep_ql = layer["indep_ql"]
    if dep_match == '':
        match_tql = ''
    else:
        match_tql = 'match ' + dep_match
    if indep_ql == '' and dep_insert == '':
        insert_tql = ''
    else:
        insert_tql = 'insert ' + indep_ql + dep_insert
    logger.info(f'\n match_tql string?-> {match_tql}')
    logger.info(f'\n insert_tql string?-> {insert_tql}')
    typeql_string = match_tql + insert_tql

    insertion_is_empty = len(insert_tql) == 0
    if insertion_is_empty:
        return None
    return typeql_string


def build_match_id_query(stix_ids: List[str]):
    get_ids_tql = 'match $id isa stix-id;'
    len_id = len(stix_ids)
    if len_id == 1:
        get_ids_tql += '$id "' + stix_ids[0] + '";'
    else:
        for index, id_l in enumerate(stix_ids):
            get_ids_tql += ' {$id "' + id_l + '";}'
            if index == len_id - 1:
                get_ids_tql += " ;"
            else:
                get_ids_tql += ' or '
    return get_ids_tql


def get_core_client(uri: str,
                    port: str):
    typedb_url = uri + ":" + port
    return TypeDB.core_client(typedb_url)

def get_data_session(core_client: TypeDBClient,
                     database: str):
    return core_client.session(database, SessionType.DATA)

def get_read_transaction(session: TypeDBSession):
    return session.transaction(TransactionType.READ)

def get_write_transaction(session: TypeDBSession):
    return session.transaction(TransactionType.WRITE)




def match_query(uri: str, port: str, database: str, query: str, data_query, **data_query_args):
    data = []
    try:
        with get_core_client(uri, port) as client:
            client_session = get_data_session(client, database)
            with client_session as session:
                read_transaction = get_read_transaction(session)
                with read_transaction as transaction:
                    answer_iterator = transaction.query().match(query)
                    data = data_query(query, answer_iterator, transaction, **data_query_args)
                    return data
    except Exception as e:
        logger.exception(e)
        raise Exception("Problem matching")

def get_all_databases(uri: str, port: str):
    client = get_core_client(uri, port)
    return client.databases().all()

def delete_database(uri: str, port: str, database: str):
    client = get_core_client(uri, port)
    if client.databases().contains(database):
       logger.info('Database ' + database + ' exists... deleting')
       client.databases().get(database).delete()
    else:
       logger.info('Database ' + database + ' does not exists... skipping')


def query_ids(query, generator, transaction, **data_query_args):
    logger.info(
        '\n\n-------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------\n' + \
        '---------------------------------------------------------------------------------------- Query ids ------------------------------------------------------------------------------\n' + \
        '-------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------\n')

    logger.info(query)

    number = 0
    ids = []
    for result in generator:
        ids.append(result.get("ids"))
        logger.info(
            '\nxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx\n' + \
            'xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx    Query IDs Concept Map ' + str(
                number) + 'xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx\n' + \
            'xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx\n')
        for concept in enumerate(result.concepts()):
            if concept[1].is_type():
                logger.info('   Type:' + concept[1].as_type().get_label())
            if concept[1].is_relation():
                logger.info('   Relation: iid ' + concept[1].as_relation().get_iid())
            if concept[1].is_attribute():
                logger.info('   Attribute iid: ' + concept[1].as_attribute().get_iid())
                logger.info('           value: ' + str(concept[1].as_attribute().get_value()))
        number = number + 1

    logger.info('\n')

    return ids

def query_id(query, generator, transaction, **data_query_args):
    logger.info(
        '\n\n-------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------\n' + \
        '---------------------------------------------------------------------------------------- Query ids ------------------------------------------------------------------------------\n' + \
        '-------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------\n')

    logger.info(query)

    number = 0
    ids = []
    for result in generator:
        ids.append(result.get("id").get_value())
        logger.info(
            '\nxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx\n' + \
            'xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx    Query ID Concept Map ' + str(
                number) + 'xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx\n' + \
            'xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx\n')
        for concept in enumerate(result.concepts()):
            if concept[1].is_type():
                logger.info('   Type:' + concept[1].as_type().get_label())
            if concept[1].is_relation():
                logger.info('   Relation: iid ' + concept[1].as_relation().get_iid())
            if concept[1].is_attribute():
                logger.info('   Attribute iid: ' + concept[1].as_attribute().get_iid())
                logger.info('           value: ' + str(concept[1].as_attribute().get_value()))
        number = number + 1

    logger.info('\n')

    return ids


def delete_layers(uri: str, port: str, database: str, instructions: Instructions):
    with get_core_client(uri, port) as client:
        client_session = get_data_session(client, database)
        with client_session as session:
            for instruction_id in instructions.get_ordered_ids():
                write_transaction = get_write_transaction(session)
                with write_transaction as transaction:
                    query = instructions.get_query_for_id(instruction_id)
                    result = delete_layer(transaction, query)
                    log_delete_layer(result, query)
                    instructions.update_delete_instruction_as_success(instruction_id)
    return instructions



def delete_layer(transaction: TypeDBTransaction, query: str):
    transaction_query: QueryManager = transaction.query()
    query_future: QueryFuture = transaction_query.delete(query)
    bi_d: BidirectionalStream = query_future.get()
    logger.info(
        '\n\n-------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------\n' + \
        '---------------------------------------------------------------------------------------- Delete Layer Query ------------------------------------------------------------------------------\n' + \
        '-------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------\n')

    logger.info(query)

    logger.info(
        '\nxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx\n' + \
        'xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx    Delete Result ' + 'xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx\n' + \
        'xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx\n')

    logger.info(str(bi_d))

    transaction.commit()


def add_layer(transaction: TypeDBTransaction, layer: str):
    transaction_query: QueryManager = transaction.query()
    query_future: Iterator[ConceptMap] = transaction_query.insert(layer)

    logger.debug('\n\n-------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------\n')
    logger.info('\n------------------------------------------------ Add Layer Query ----------------------------------------------\n')
    logger.debug('-------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------\n')

    logger.info(layer)

    number = 0
    logger.info('\nxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx    Add Layer Response     xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx\n')
    logger.info(f'insert_iterator response ->\n{query_future}')
    for result in query_future:
        logger.debug('\nxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx\n' )
        logger.debug('\nxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx    Add Layer Concept Map ' + str(number) +      'xxxxxxxxxxxxxxxxxxxxxx\n')
        logger.debug('\nxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx\n')

        logger.debug(f'typedb response ->\n{result}')

        for concept in enumerate(result.concepts()):
            if concept[1].is_type():
                logger.debug('   Type:' + concept[1].as_type().get_label())
            if concept[1].is_relation():
                logger.debug('   Relation: iid ' + concept[1].as_relation().get_iid() )
            if concept[1].is_attribute():
                logger.debug('   Attribute iid: ' + concept[1].as_attribute().get_iid())
                logger.debug('           value: ' + str(concept[1].as_attribute().get_value()))
        number = number + 1

    logger.info('\n\n')

    transaction.commit()


def add_instructions_to_typedb(uri: str, port: str, database: str, instructions: Instructions):
    try:
        with get_core_client(uri, port) as client:
            client_session = get_data_session(client, database)
            with client_session as session:
                for instruction_id in instructions.get_ordered_ids():
                    if instructions.not_allow_insertion(instruction_id):
                        continue
                    write_transaction = get_write_transaction(session)
                    with write_transaction as transaction:
                        query = instructions.get_query_for_id(instruction_id)
                        add_layer(transaction, query)
                        instructions.update_instruction_as_success(instruction_id)
    except Exception as e:
        traceback_str = traceback.format_exc()
        instructions.update_instruction_as_error(instruction_id, traceback_str)
    return instructions





"""Python Mitre ATT&CK Stix Class Definitions """
from stixorm.module.typedb_lib.factories.mappings_factory import get_mapping_factory_instance

"""Python Mitre ATT&CK Stix Class Definitions """
import json
import pathlib
import warnings
from collections import OrderedDict

from stix2.exceptions import (
    PropertyPresenceError, STIXDeprecationWarning, )
from stix2.properties import (
    BooleanProperty, ExtensionsProperty, IDProperty, IntegerProperty, ListProperty,
    OpenVocabProperty, ReferenceProperty, StringProperty,
    TimestampProperty, TypeProperty, EmbeddedObjectProperty, ObservableProperty
)
from stix2.utils import NOW, _get_dict
from stix2.markings import _MarkingsMixin
from stix2.markings.utils import check_tlp_marking
from stix2.v21.base import _DomainObject, _STIXBase21, _RelationshipObject
from stix2.v21.common import (
    ExternalReference, GranularMarking, KillChainPhase,
    MarkingProperty, TLPMarking, StatementMarking,
)
from stix2.v21.vocab import (
    ATTACK_MOTIVATION, ATTACK_RESOURCE_LEVEL, IMPLEMENTATION_LANGUAGE, MALWARE_CAPABILITIES, MALWARE_TYPE,
    PROCESSOR_ARCHITECTURE, TOOL_TYPE, IDENTITY_CLASS, INDUSTRY_SECTOR, REPORT_TYPE
)

import logging

from stixorm.module.definitions.property_definitions import ThreatReference, ThreatExtensionsProperty


logger = logging.getLogger(__name__)

valid_obj =  get_mapping_factory_instance().get_all_types()


class Note(_DomainObject):
    """For more detailed information on this object's properties, see
    `the STIX 2.1 specification <https://docs.oasis-open.org/cti/stix/v2.1/os/stix-v2.1-os.html#_gudodcg1sbb9>`__.
    """

    _type = 'note'
    _properties = OrderedDict([
        ('type', TypeProperty(_type, spec_version='2.1')),
        ('spec_version', StringProperty(fixed='2.1')),
        ('id', IDProperty(_type, spec_version='2.1')),
        ('created_by_ref', ReferenceProperty(valid_types='identity', spec_version='2.1')),
        ('created', TimestampProperty(default=lambda: NOW, precision='millisecond', precision_constraint='min')),
        ('modified', TimestampProperty(default=lambda: NOW, precision='millisecond', precision_constraint='min')),
        ('abstract', StringProperty()),
        ('content', StringProperty(required=True)),
        ('authors', ListProperty(StringProperty)),
        ('object_refs', ListProperty(ThreatReference(valid_types=valid_obj, spec_version='2.1'), required=True)),
        ('revoked', BooleanProperty(default=lambda: False)),
        ('labels', ListProperty(StringProperty)),
        ('confidence', IntegerProperty()),
        ('lang', StringProperty()),
        ('external_references', ListProperty(ExternalReference)),
        ('object_marking_refs', ListProperty(ReferenceProperty(valid_types='marking-definition', spec_version='2.1'))),
        ('granular_markings', ListProperty(GranularMarking)),
        ('extensions', ExtensionsProperty(spec_version='2.1')),
    ])


class ObservedData(_DomainObject):
    """For more detailed information on this object's properties, see
    `the STIX 2.1 specification <https://docs.oasis-open.org/cti/stix/v2.1/os/stix-v2.1-os.html#_p49j1fwoxldc>`__.
    """

    _type = 'observed-data'
    _properties = OrderedDict([
        ('type', TypeProperty(_type, spec_version='2.1')),
        ('spec_version', StringProperty(fixed='2.1')),
        ('id', IDProperty(_type, spec_version='2.1')),
        ('created_by_ref', ReferenceProperty(valid_types='identity', spec_version='2.1')),
        ('created', TimestampProperty(default=lambda: NOW, precision='millisecond', precision_constraint='min')),
        ('modified', TimestampProperty(default=lambda: NOW, precision='millisecond', precision_constraint='min')),
        ('first_observed', TimestampProperty(required=True)),
        ('last_observed', TimestampProperty(required=True)),
        ('number_observed', IntegerProperty(min=1, max=999999999, required=True)),
        ('objects', ObservableProperty(spec_version='2.1')),
        ('object_refs', ListProperty(ThreatReference(valid_types=valid_obj, spec_version='2.1'), required=True)),
        ('revoked', BooleanProperty(default=lambda: False)),
        ('labels', ListProperty(StringProperty)),
        ('confidence', IntegerProperty()),
        ('lang', StringProperty()),
        ('external_references', ListProperty(ExternalReference)),
        ('object_marking_refs', ListProperty(ReferenceProperty(valid_types='marking-definition', spec_version='2.1'))),
        ('granular_markings', ListProperty(GranularMarking)),
        ('extensions', ExtensionsProperty(spec_version='2.1')),
    ])

    def __init__(self, *args, **kwargs):

        if "objects" in kwargs:
            warnings.warn(
                "The 'objects' property of observed-data is deprecated in "
                "STIX 2.1.",
                STIXDeprecationWarning,
            )

        super(ObservedData, self).__init__(*args, **kwargs)

    def _check_object_constraints(self):
        super(ObservedData, self)._check_object_constraints()

        first_observed = self.get('first_observed')
        last_observed = self.get('last_observed')

        if first_observed and last_observed and last_observed < first_observed:
            msg = "{0.id} 'last_observed' must be greater than or equal to 'first_observed'"
            raise ValueError(msg.format(self))

        self._check_mutually_exclusive_properties(
            ["objects", "object_refs"],
        )

class Incident(_DomainObject):
    """For more detailed information on this object's properties, see
    `the STIX 2.1 specification <https://docs.oasis-open.org/cti/stix/v2.1/os/stix-v2.1-os.html#_sczfhw64pjxt>`__.
    """

    _type = 'incident'
    _properties = OrderedDict([
        ('type', TypeProperty(_type, spec_version='2.1')),
        ('spec_version', StringProperty(fixed='2.1')),
        ('id', IDProperty(_type, spec_version='2.1')),
        ('created_by_ref', ReferenceProperty(valid_types='identity', spec_version='2.1')),
        ('created', TimestampProperty(default=lambda: NOW, precision='millisecond', precision_constraint='min')),
        ('modified', TimestampProperty(default=lambda: NOW, precision='millisecond', precision_constraint='min')),
        ('name', StringProperty(required=True)),
        ('description', StringProperty()),
        ('kill_chain_phases', ListProperty(KillChainPhase)),
        ('revoked', BooleanProperty(default=lambda: False)),
        ('labels', ListProperty(StringProperty)),
        ('confidence', IntegerProperty()),
        ('lang', StringProperty()),
        ('external_references', ListProperty(ExternalReference)),
        ('object_marking_refs', ListProperty(ReferenceProperty(valid_types='marking-definition', spec_version='2.1'))),
        ('granular_markings', ListProperty(GranularMarking)),
        ('extensions', ThreatExtensionsProperty(spec_version='2.1')),
    ])

valid_obj =  get_mapping_factory_instance().get_all_types()
class Report(_DomainObject):
    """For more detailed information on this object's properties, see
    `the STIX 2.1 specification <https://docs.oasis-open.org/cti/stix/v2.1/os/stix-v2.1-os.html#_n8bjzg1ysgdq>`__.
    """

    _type = 'report'
    _properties = OrderedDict([
        ('type', TypeProperty(_type, spec_version='2.1')),
        ('spec_version', StringProperty(fixed='2.1')),
        ('id', IDProperty(_type, spec_version='2.1')),
        ('created_by_ref', ReferenceProperty(valid_types='identity', spec_version='2.1')),
        ('created', TimestampProperty(default=lambda: NOW, precision='millisecond', precision_constraint='min')),
        ('modified', TimestampProperty(default=lambda: NOW, precision='millisecond', precision_constraint='min')),
        ('name', StringProperty(required=True)),
        ('description', StringProperty()),
        ('report_types', ListProperty(OpenVocabProperty(REPORT_TYPE))),
        ('published', TimestampProperty(required=True)),
        ('object_refs', ListProperty(ThreatReference(valid_types=valid_obj, spec_version='2.1'), required=True)),
        ('revoked', BooleanProperty(default=lambda: False)),
        ('labels', ListProperty(StringProperty)),
        ('confidence', IntegerProperty()),
        ('lang', StringProperty()),
        ('external_references', ListProperty(ExternalReference)),
        ('object_marking_refs', ListProperty(ReferenceProperty(valid_types='marking-definition', spec_version='2.1'))),
        ('granular_markings', ListProperty(GranularMarking)),
        ('extensions', ExtensionsProperty(spec_version='2.1')),
    ])



import logging
from typing import List

from stixorm.module.definitions.property_definitions import get_libraries
from stixorm.module.typedb_lib.factories.definition_factory import get_definition_factory_instance, DomainDefinition
from stixorm.module.typedb_lib.factories.import_type_factory import ImportTypeFactory, ImportType
from stixorm.module.typedb_lib.factories.process_map_factory import ProcessMapFactory

logger = logging.getLogger(__name__)



##############################################################
#  1.) Default Import Type at only Stix Objects, all else False
############################################################


import_type_factory = ImportTypeFactory.get_import_type_factory()
default_import_type = import_type_factory.get_default_import()


def authorised_mappings(import_type: ImportType=default_import_type):
    auth = {}
    auth["reln_name"] = {}
    auth["reln"] = {}
    auth["tql_types"] = {}
    auth["types"] = {}
    auth["is_lists"] = {}
    definition_factory = get_definition_factory_instance()
    auth_domains: List[DomainDefinition] = definition_factory.get_definitions_for_import(import_type)


    dom= get_libraries()
    # initialise authorisation object, for documentation purposes
    auth["reln_name"]["embedded_relations"] = []
    auth["reln_name"]["standard_relations"] = []
    auth["reln_name"]["list_of_objects"] = []
    auth["reln_name"]["key_value_relations"] = []
    auth["reln_name"]["extension_relations"] = []
    auth["reln_name"]["relations_sro_roles"] = []
    auth["reln"]["embedded_relations"] = []
    auth["reln"]["standard_relations"] = []
    auth["reln"]["list_of_objects"] = []
    auth["reln"]["key_value_relations"] = []
    auth["reln"]["extension_relations"] = []
    auth["reln"]["relations_sro_roles"] = []
    auth["tql_types"]["embedded_relations"] = []
    auth["tql_types"]["standard_relations"] = []
    auth["tql_types"]["list_of_objects"] = []
    auth["tql_types"]["key_value_relations"] = []
    auth["tql_types"]["extension_relations"] = []
    auth["tql_types"]["relations_sro_roles"] = []
    auth["tql_types"]["sdo"] = []
    auth["tql_types"]["sro"] = []
    auth["tql_types"]["sco"] = []
    auth["tql_types"]["sub"] = []
    auth["tql_types"]["meta"] = []
    auth["types"]["sdo"] = []
    auth["types"]["sro"] = []
    auth["types"]["sco"] = []
    auth["types"]["sub"] = []
    auth["types"]["meta"] = []
    auth["is_lists"]["sdo"] = {}
    auth["is_lists"]["sro"] = {}
    auth["is_lists"]["sco"] = {}
    auth["is_lists"]["sub"] = {}
    auth["is_lists"]["meta"] = {}
    auth["sub_objects"] = {}
    auth["objects"] = {}
    auth["conv"] = {}
    auth["conv"]["sdo"] = []
    auth["conv"]["sro"] = []
    auth["conv"]["sco"] = []
    auth["conv"]["sub"] = []
    auth["conv"]["meta"] = []
    auth["classes"] = {}
    auth["classes"]["sdo"] = {}
    auth["classes"]["sro"] = {}
    auth["classes"]["sco"] = {}
    auth["classes"]["sub"] = {}
    auth["classes"]["meta"] = {}

    process_map_factory = ProcessMapFactory.process_map_factory()
    process_maps = process_map_factory.all_process_maps()

    for j, domain in enumerate(auth_domains):
        for process in process_maps:
            name = process.name
            keys = process.keys
            matches = process.match
            conds = process.cond
            if name == "reln_name":
                #logger.debug("----------- reln_name ------------")
                for i, key in enumerate(keys):
                    if domain.contains_mapping(matches[i]):
                        #logger.debug(f'Auth Loading: domain->{dom[j]}, name->{name}, key->{keys[i]}, match->{matches[i]}, cond->{conds[i]}')
                        value_list = [x[conds[i]] for x in domain.get_mapping(matches[i])]
                        auth[name][key].extend(value_list)
            elif name == "reln":
                #logger.debug("--------- reln--------------")
                for i, key in enumerate(keys):
                    if domain.contains_mapping(matches[i]):
                        #logger.debug(f'Auth Loading: domain->{dom[j]}, name->{name}, key->{key}, match->{matches[i]}')
                        value_list =  domain.get_mapping(matches[i])
                        auth[name][key].extend(value_list)
            elif name == "tql_types1":
                #logger.debug("--------- reln--------------")
                for i, key in enumerate(keys):
                    if domain.contains_mapping(matches[i]):
                        logger.debug(f'\nAuth Loading: domain->{dom[j]}, name->{name}, key->{key}, match->{matches[i]}')
                        value_list = [x[conds[i]] for x in domain.get_mapping(matches[i])]
                        logger.debug(f'value list -> {value_list}')
                        auth["tql_types"][key].extend(value_list)
            elif name == "tql_types":
                #logger.debug("---------- tql_types -------------")
                for i, key in enumerate(keys):
                    if domain.contains_mapping("object_conversion"):
                        #logger.debug(f'Auth Loading: domain->{dom[j]}, name->{name}, key->{keys[i]}, match->{matches[i]}, cond->{conds[i]}')
                        value_list_type = [x["type"] for x in domain.get_mapping(matches[i]) if x["object"] == conds[i]]
                        value_list_typeql = [x["typeql"] for x in domain.get_mapping(matches[i]) if x["object"] == conds[i]]
                        logger.debug(f' value_list_type -> {value_list_type}\n\n value_list_typeql -> {value_list_typeql}')
                        auth["tql_types"][key].extend(value_list_typeql)
                        auth["types"][key].extend(value_list_type)

                #auth["tql_types"]["meta"] = stix_models["mappings"]["types_meta"]
            elif name == "is_lists":
                #logger.debug("--------- is_lists --------------")
                for i, key in enumerate(keys):
                    if domain.contains_mapping(matches[i]):
                        #logger.debug(f'Auth Loading: domain->{dom[j]}, name->{name}, key->{keys[i]}, match->{matches[i]}, cond->{conds[i]}')
                        value_dict = domain.get_mapping(matches[i])
                        auth[name][key].update(value_dict)
            elif name == "direct":
                #logger.debug("-------- direct ---------------")
                for i, key in enumerate(keys):
                    if domain.does_property_contain_values(matches[i]):
                        #logger.debug(f'Auth Loading: domain->{dom[j]}, name->{name}, key->{keys[i]}, match->{matches[i]}')
                        value_dict = domain.get_property_values(matches[i])
                        auth[key].update(value_dict)
            elif name == "conv":
                #logger.debug("-------- conv ---------------")
                for i, key in enumerate(keys):
                    if domain.contains_mapping("object_conversion"):
                        #logger.debug(f'Auth Loading: domain->{dom[j]}, name->{name}, key->{keys[i]}, match->{matches[i]}, cond->{conds[i]}')
                        value_list = [x for x in domain.get_mapping("object_conversion") if x["object"] == conds[i]]
                        auth[name][key].extend(value_list)
            elif name == "classes":
                #logger.debug("-------- conv ---------------")
                for i, key in enumerate(keys):
                    if domain.does_classes_contain_values(key):
                        #logger.debug(f'Auth Loading: domain->{dom[j]}, name->{name}, key->{keys[i]}, match->{matches[i]}, cond->{conds[i]}')
                        value_dict = domain.get_classes_property_values(key)
                        auth[name][key].update(value_dict)

            else:
                pass


    # finally add the import type to the auth object
    auth.update(ImportTypeFactory.convert_to_dict(import_type))

    return auth




"""Python STIX2 TypeDB Source/Sink"""
import os.path
import pathlib
import traceback
from dataclasses import dataclass
from typedb.client import *
from stixorm.module.orm.import_objects import raw_stix2_to_typeql
from stixorm.module.orm.delete_object import delete_stix_object, add_delete_layers
from stixorm.module.orm.export_object import convert_ans_to_stix
from stixorm.module.parsing.parse_objects import parse
from .authorise import authorised_mappings, import_type_factory
from .initialise import setup_database, load_schema, load_markings
import networkx as nx
from stix2 import v21
from stix2.base import _STIXBase
from stix2.datastore import (
    DataSink, DataSource, )
from stix2.datastore.filters import FilterSet

import logging

from stixorm.module.typedb_lib.handlers import handle_result
from stixorm.module.typedb_lib.logging import log_delete_instruction, log_delete_instruction_update_layer, log_delete_layers
from stixorm.module.typedb_lib.queries import delete_database, match_query, query_ids, delete_layers, build_match_id_query,\
    build_insert_query, query_id, add_instructions_to_typedb
from stixorm.module.typedb_lib.file import write_to_file
from stixorm.module.typedb_lib.instructions import Instructions, Status, AddInstruction, TypeQLObject, Result
from stixorm.module.typedb_lib.factories.import_type_factory import ImportType, ImportTypeFactory
from stixorm.module.typedb_lib.factories.auth_factory import get_auth_factory_instance

# logging.basicConfig(level=logging.INFO, format='[%(asctime)s] %(levelname)s [%(name)s:%(lineno)s] %(message)s')

logger = logging.getLogger(__name__)


@dataclass
class TransactionObject:
    transaction: TypeDBTransaction
    session: TypeDBSession



class TypeDBSink(DataSink):
    """Interface for adding/pushing STIX objects to TypeDB.

    Can be paired with a TypeDBSource, together as the two
    components of a TypeDBStore.

    Args:
        - connection is a dict, containing:
            - uri (str): URI to TypeDB.
            - port (int): Port to TypeDB.
            - db (str): Name of TypeDB database.
            - user (str): Username for TypeDB, if cluster, otherwise None
            - password (str): Password for TypeDB, if cluster, otherwise None
        - clear (bool): If True, clear the TypeDB before adding objects.
        - import_type (str): It forces the parser to use either the stix2.1, or mitre att&ck

    """

    def __init__(self,
                 connection: Dict[str, str],
                 clear=False,
                 import_type: Optional[ImportType]=None,
                 schema_path: Optional[str] = None,
                 strict_failure: bool = False, **kwargs):
        super(TypeDBSink, self).__init__()
        logger.debug(f'TypeDBSink: {connection}')

        assert connection["uri"] is not None
        assert connection["port"] is not None
        assert connection["database"] is not None

        self._stix_connection = connection
        self.uri: str = connection["uri"]
        self.port: str = connection["port"]
        self.database: str = connection["database"]
        self.user: str = connection["user"]
        self.password: str = connection["password"]
        self.clear: bool = clear
        self.strict_failure = strict_failure

        self.schema_path = schema_path
        self.import_type: ImportType = import_type

        self.__initialise()



    def __initialise(self):
        assign_result = self.__assign_schemas()
        handle_result(assign_result, "assign result", self.strict_failure)

        assign_import_result = self.__assign_import_type()
        handle_result(assign_import_result, "assign import result", self.strict_failure)

        # Validate database can be connected
        validate_connection = self.__validate_connect_to_db()
        handle_result(validate_connection, "validate connection result", self.strict_failure)

        # 1. Setup database
        setup_database(self._stix_connection, self.clear)

        # 2. Load the Schema's
        # A. Load the Stix Schema
        schema_result = self.__load_stix_schema()
        handle_result(schema_result, "history schema result", self.strict_failure)
        # B. Load Stix Rules Schema
        rules_result = self.__load_stix_rules()
        handle_result(rules_result, "history stix rules result", self.strict_failure)
        # C. Load the Attack Schema
        attack_result = self.__load_attack_schema()
        handle_result(attack_result, "history attack result", self.strict_failure)
        # D. Load the OS-Threat Schema
        os_threat_result = self.__load_os_threat_schema()
        handle_result(os_threat_result, "history os threat result", self.strict_failure)

        # 3. Load the Objects
        # Still to do

    def __validate_connect_to_db(self):
        logger.debug("Attempting DB Connection")
        self.__get_core_client()
        logger.debug("DB Connection Successful")


    def __load_attack_schema(self):
        if self.clear and self.import_type.ATTACK:
            logger.debug("ATT&CK")
            load_schema(self._stix_connection, str(self.cti_schema_attack), "ATT&CK Schema")
            logger.info("we have loaded ATT&CK schema")
        else:
            logger.debug("ignoring history ATT&CK schema")


    def __load_os_threat_schema(self):
        if self.clear and self.import_type.os_threat:
            logger.debug("os-threat")
            load_schema(self._stix_connection, str(self.cti_schema_os_threat), "os-threat Schema ")
            logger.info("we have loaded os-threat schema")
        else:
            logger.debug("ignoring history  os hunt")


    def __load_stix_rules(self):
        if self.clear and self.import_type.rules:
            logger.debug("rules")
            load_schema(self._stix_connection, str(self.cti_schema_stix_rules), "Stix 2.1 Rules")
            logger.info("we have loaded Stix rules")
        else:
            logger.debug("ignoring check of stix rules")


    def __load_stix_schema(self):
        if self.clear:
            load_schema(self._stix_connection, str(self.cti_schema_stix), "Stix 2.1 Schema ")
            self.loaded = load_markings(self._stix_connection)
            logger.info("we have loaded Stix schema")
        else:
            logger.debug("ignoring history stix schema")


    def __assign_schemas(self):
        if self.schema_path is None:
             self.schema_path = str(pathlib.Path(__file__).parent)

        # If relative paths are used it will depend upon the entry point i.e. working directory will need to be same level as typedb.py
        self.cti_schema_stix = pathlib.Path(self.schema_path).joinpath("definitions/stix21/schema/cti-schema-v2.tql")
        assert os.path.isfile(self.cti_schema_stix)
        self.cti_schema_stix_rules = pathlib.Path(self.schema_path).joinpath("definitions/stix21/schema/cti-rules.tql")
        assert os.path.isfile(self.cti_schema_stix_rules)
        self.cti_schema_os_threat = pathlib.Path(self.schema_path).joinpath("definitions/os_threat/schema/cti-os-threat.tql")
        assert os.path.isfile(self.cti_schema_os_threat)
        self.cti_schema_os_hunt = pathlib.Path(self.schema_path).joinpath("definitions/os_threat/schema/cti-os-hunt.tql")
        assert os.path.isfile(self.cti_schema_os_hunt)
        self.cti_schema_attack = pathlib.Path(self.schema_path).joinpath("definitions/attack/schema/cti-attack.tql")
        assert os.path.isfile(self.cti_schema_attack)
        # if self.schema_path is None:
        #     self.schema_path = str(pathlib.Path.parent)
        #
        # self.cti_schema_path = pathlib.Path(self.schema_path).joinpath("stix/schema/cti-schema-v2.tql")
        # assert self.cti_schema_path.is_file(), "The schema does not exist: " + str(self.cti_schema_path)
        #
        # self.cti_schema_rules_path = pathlib.Path(self.schema_path).joinpath("stix/schema/cti-rules.tql")
        # assert self.cti_schema_rules_path.is_file(), "The schema does not exist: " + str(self.cti_schema_rules_path)

    def __assign_import_type(self):
        if self.import_type is None:
            self.import_type = import_type_factory.get_default_import()

    @property
    def stix_connection(self):
        return self._stix_connection

    def clear_db(self):

        result = delete_database(self.uri, self.port, self.database)
        logger.debug("Successfully cleared database")


    def __filter_markings(self, stix_ids: List[StringAttribute]) -> List[str]:
        marking = ["marking-definition--613f2e26-407d-48c7-9eca-b8e91df99dc9",
                   "marking-definition--34098fce-860f-48ae-8e50-ebd3cc5e41da",
                   "marking-definition--f88d31f6-486f-44da-b317-01333bde0b82",
                   "marking-definition--5e57c739-391a-4eb3-b6be-7d15ca92d5ed"]

        filtered_list = list(filter(lambda x: x.get_value() not in marking, stix_ids))
        return self.__string_attibute_to_string(filtered_list)

    def __string_attibute_to_string(self,
                                    string_attributes: List[StringAttribute]):
        return [stix_id.get_value() for stix_id in string_attributes]

    def __query_stix_ids(self):
        get_ids_tql = 'match $ids isa stix-id;'
        data_query = query_ids
        query_data = match_query(self.uri,
                                 self.port,
                                 self.database,
                                 get_ids_tql,
                                 data_query)

        return query_data


    def get_stix_ids(self):
        """ Get all the stix-ids in a database, should be moved to DataSource object

        Returns:
            id_list : list of the stix-ids in the database
        """
        stix_ids_query = self.__query_stix_ids()

        result = self.__filter_markings(stix_ids_query)
        return result


    def __retrieve_stix_id(self,
                           stix_id: str):
        type_db_source = self.__get_source_client()
        return type_db_source.get(stix_id)


    def __delete_instruction(self,
                             stixid: str):

        stix_obj = self.__retrieve_stix_id(stixid)

        dep_match, dep_insert, indep_ql, core_ql, dep_obj = raw_stix2_to_typeql(stix_obj, self.import_type)
        del_match, del_tql = delete_stix_object(stix_obj, dep_match, dep_insert, indep_ql, core_ql, self.import_type)
        dep_obj["delete"] = del_match + '\n' + del_tql

        log_delete_instruction(del_match, dep_insert, indep_ql, dep_obj, del_match, del_tql)
        if del_match == '' and del_tql == '':
            return None
        else:
            return dep_obj


    def __update_delete_layers(self,
                               layers,
                               indexes,
                               missing,
                               dep_obj):
        if dep_obj is None:
            return layers, indexes, missing
        if len(layers) == 0:
            missing = dep_obj['dep_list']
            indexes.append(dep_obj['id'])
            layers.append(dep_obj)
        else:
            layers, indexes, missing = add_delete_layers(layers, dep_obj, indexes, missing)
        return layers, indexes, missing


    def __retrieve_delete_instructions(self,
                                       stixids: List[str]) -> Instructions:

        layers = []
        indexes = []
        missing = []

        instructions = Instructions()

        for stixid in stixids:
            del_result = self.__delete_instruction(stixid)
            layers, indexes, missing = self.__update_delete_layers(layers, indexes, missing, del_result)

        for layer in layers:
            instructions.insert_delete_instruction(layer['id'], layer)
        return instructions


    def __order_delete_instructions(self,
                                    delete_instructions: Instructions):
        layer = {}
        layer['delete'] = 'match $a isa attribute; not { $b isa thing; $b has $a;}; delete $a isa attribute;'
        delete_instructions.insert_delete_instruction(
            "cleanup-1", layer
        )
        delete_instructions.insert_delete_instruction(
            "cleanup-2", layer
        )
        return delete_instructions

    def delete(self, stixid_list: List[str]) -> Instructions:
        """ Delete a list of STIX objects from the typedb_lib server. Must include all related objects and relations

        Args:
            stixid_list (): The list of Stix-id's of the object's to delete
        """

        delete_instruction_result = self.__retrieve_delete_instructions(stixid_list)
        order_instruction_result = self.__order_delete_instructions(delete_instruction_result)
        delete_from_database_result = delete_layers(self.uri,
                                                    self.port,
                                                    self.database,
                                                    order_instruction_result)

        instructions = delete_from_database_result

        return instructions.convert_to_result()



    def __get_core_client(self) -> TypeDBClient:
        typedb_url = self.uri + ":" + self.port
        return TypeDB.core_client(typedb_url)


    def __get_source_client(self):
        connection = {'uri': self.uri,
                      'port': self.port,
                      'database': self.database,
                      'user': self.user,
                      'password': self.password}

        return TypeDBSource(connection, self.import_type)




    def __generate_typeql_object(self, stix_dict: dict) -> TypeQLObject:

        logger.debug(f"\n================================================================\nim about to parse \n")
        stix_obj = parse(stix_dict, False, self.import_type)
        logger.debug(f'\n-------------------------------------------------------------\n i have parsed\n')
        dep_match, dep_insert, indep_ql, core_ql, dep_obj = raw_stix2_to_typeql(stix_obj, self.import_type)
        logger.debug(f'\ndep_match {dep_match} \ndep_insert {dep_insert} \nindep_ql {indep_ql} \ncore_ql {core_ql}')
        typeql_obj = TypeQLObject(
            dep_match=dep_match,
            dep_insert=dep_insert,
            indep_ql=indep_ql,
            core_ql=core_ql,
            dep_list=dep_obj.get('dep_list', [])
        )

        return typeql_obj


    def __generate_instructions(self,
                                obj_list) -> Instructions:
        instructions = Instructions()
        for stix_dict in obj_list:
            try:
                typeql_object_result = self.__generate_typeql_object(stix_dict)

                typeql_object: TypeQLObject = typeql_object_result
                instructions.insert_add_instruction(stix_dict['id'], typeql_object)
            except Exception as e:
                logging.exception(e)
                logging.info("Error generating instructions for " + stix_dict['id'])
                instructions.insert_add_instruction(stix_dict['id'], None)
                traceback_str = traceback.format_exc()
                instructions.update_instruction_as_error(stix_dict['id'], traceback_str)


        return instructions


    def __create_instruction_dependency_graph(self,
                                              instructions: Instructions):
        directed_graph = nx.DiGraph()
        instruction: AddInstruction
        for instruction in instructions.instructions.values():
            try:
                if instruction.status in [Status.ERROR]:
                    logging.debug("Skipping error " + instruction.id)
                    continue

                dependencies = instruction.typeql_obj.dep_list

                this_node = instruction.id
                if not directed_graph.has_node(this_node):
                    # logging.info("Already has node id " + this_node)
                    # logging.info("Inserting dependency node " + this_node)
                    directed_graph.add_node(this_node)

                for dependency_node in dependencies:
                    if not directed_graph.has_node(dependency_node):
                       # logging.info("Already has dependency node id " + this_node)

                       # logging.info("Dependency node does not exist id " + this_node)
                        directed_graph.add_node(this_node)
                    directed_graph.add_edge(dependency_node, this_node)
                instructions.add_dependencies(directed_graph)
            except Exception as e:
                logging.error("Error creating dependency graph for " + instruction.id)
                logging.error(e)
                traceback_str = traceback.format_exc()
                instructions.update_instruction_as_error(instruction.id, traceback_str)

        return instructions



    def __generate_queries(self,
                           instructions: Instructions):
        result = instructions.create_insert_queries(build_insert_query)
        return result


    def __check_missing_dependencies(self,
                                     instructions: Instructions):
        missing_ids_from_tree = instructions.missing_dependency_ids()

        result = []
        for missing_ids in missing_ids_from_tree[:500]:
            query_result = build_match_id_query(missing_ids)

            data_result = match_query(uri=self.uri,
                                      port=self.port,
                                      database=self.database,
                                      query=query_result,
                                      data_query=query_id,
                                      import_type=None)
            result = result + data_result


        missing_ids_found_in_db = result
        # ids with no record in db and in dependency tree
        ids_missing = list(set(missing_ids_from_tree) - set(missing_ids_found_in_db))
        instructions.register_missing_dependencies(ids_missing)

        return instructions



    def __reorder_instructions(self,
                               instructions: Instructions):
        try:
            order = list(nx.topological_sort(instructions.dependencies))
            instructions.add_insertion_order(order)
        except Exception as e:
            logging.exception(e)
        return instructions




    def add(self, stix_data: Optional[List[dict]] = None) -> List[Result]:
        """Add STIX objects to the typedb_lib server.
            1. Gather objects into a list
            2. For each object
                a. get raw stix to tql
                b. add object to an ordered list
                c. return the ordered list
            3. Add each object in the ordered list
        Args:
            stix_data (STIX object OR Bundle OR dict OR list): valid STIX 2.1 content
                in a STIX object (or list of), dict (or list of), or a STIX 2.1
                json encoded string.
            import_type (dict): It forces the parser to use either the stix2.1,
                or the mitre attack typeql description. Values can be either:
                        - "STIX21"
                        - "mitre"
        Note:
            ``stix_data`` can be a Bundle object, but each object in it will be
            saved separately; you will be able to retrieve any of the objects
            the Bundle contained, but not the Bundle itself.
        """
        logger.debug("1. starting in add")
        obj_result = self._gather_objects(stix_data)

        generate_instructions_result = self.__generate_instructions(obj_result)
        logger.info("\n##########################################################################################################################################################\n")
        #logger.info(f"generate instructions is {generate_instructions_result}")
        instruction_dependency_graph_result =  self.__create_instruction_dependency_graph(generate_instructions_result)
        check_missing_dependency_result = self.__check_missing_dependencies(instruction_dependency_graph_result)

        instructions: Instructions = check_missing_dependency_result
        if instructions.exist_missing_dependencies():
            return instructions.convert_to_result()
        if instructions.exist_cyclical_ids():
            instructions.register_cyclical_dependencies()
            return instructions.convert_to_result()

        reorder_result = self.__reorder_instructions(instructions)

        queries_result = self.__generate_queries(reorder_result)

        add_to_database_result = add_instructions_to_typedb(self.uri,
                                                            self.port,
                                                            self.database,
                                                            reorder_result)

        instructions = add_to_database_result

        return instructions.convert_to_result()



    def _gather_objects(self, stix_data):
        """
          the details for the add details, checking what import_type of data object it is
        """
        #logger.debug(f" gethering ...{stix_data}")
        #logger.debug('----------------------------------------')
        #logger.debug(f'going into separate objects function {stix_data}')
        #logger.debug('-----------------------------------------------------')

        if isinstance(stix_data, (v21.Bundle)):
            logger.debug(f'isinstance Bundle')
            # recursively add individual STIX objects
            logger.debug(f'obects are {stix_data["objects"]}')
            return stix_data.get("objects", [])


        elif isinstance(stix_data, _STIXBase):
            logger.debug("base")
            logger.debug(f'isinstance _STIXBase')
            temp_list = []
            temp_list.append(stix_data)
            return temp_list

        elif isinstance(stix_data, (str, dict)):
            if stix_data.get("type", '') == 'bundle':
                return stix_data.get("objects", [])
            else:
                logger.debug("dcit")
                logger.debug(f'isinstance dict')
                temp_list = []
                temp_list.append(stix_data)
                return temp_list

        elif isinstance(stix_data, list):
            item_list = []
            for item in stix_data:
                if item.get("type", '') == 'bundle':
                    item_list = item_list + item.get("objects", [])
                else:
                    item_list.append(item)

            logger.debug(f'isinstance list')
            # recursively add individual STIX objects
            return item_list

        else:
            raise TypeError(
                "stix_data must be a STIX object (or list of), "
                "JSON formatted STIX (or list of), "
                "or a JSON formatted STIX bundle",
            )


class TypeDBSource(DataSource):
    """Interface for searching/retrieving STIX objects from a TypeDB Database.

    Can be paired with a TypeDBSink, together as the two
    components of a TypeDBStore.

    Args:
        - connection is a dict, containing:
            - uri (str): URI to TypeDB.
            - port (int): Port to TypeDB.
            - db (str): Name of TypeDB database.
            - user (str): Username for TypeDB, if cluster, otherwise None
            - password (str): Password for TypeDB, if cluster, otherwise None
        - import_type (str): It forces the parser to use either the stix2.1, or mitre att&ck

    """

    def __init__(self, connection: Dict[str, str], import_type: Optional[ImportType]=None, **kwargs):
        super(TypeDBSource, self).__init__()
        logger.debug(f'TypeDBSource: {connection}')

        assert connection["uri"] is not None
        assert connection["port"] is not None
        assert connection["database"] is not None
        assert import_type is None or isinstance(import_type, ImportType)

        self._stix_connection = connection
        self.uri = connection["uri"]
        self.port = connection["port"]
        self.database = connection["database"]
        self.user = connection["user"]
        self.password = connection["password"]
        self.import_type: ImportType = self.__default_import_type() if import_type is None else import_type

    def __default_import_type(self):
        return ImportTypeFactory.get_default_import()

    @property
    def stix_connection(self):
        return self._stix_connection


    def __retrieve_stix_object(self,
                               stix_id: str):
        logger.debug(f'__retrieve_stix_object: {stix_id}')
        obj_var, type_ql = get_embedded_match(stix_id, self.import_type)
        query = 'match ' + type_ql
        logger.debug(f'query is {query}')

        data = match_query(uri=self.uri,
                           port=self.port,
                           database=self.database,
                           query=query,
                           data_query=convert_ans_to_stix,
                           import_type=self.import_type)

        logger.debug(f'data is -> {data}')
        stix_obj = parse(data=data, allow_custom=False, import_type=self.import_type)

        # result = write_to_file("stixorm/module/orm/export_final.json", stix_obj)
        # if not is_successful(result):
        #     logging.exception("\n".join(traceback.format_exception(result.failure())))
        #     logger.error(str(result.failure()))

        return stix_obj

    def get(self, stix_id: str, _composite_filters=None):
        """Retrieve STIX object from file directory via STIX ID.

        Args:
            stix_id (str): The STIX ID of the STIX object to be retrieved.
            _composite_filters (FilterSet): collection of filters passed from the parent
                CompositeDataSource, not user supplied

        Returns:
            (STIX object): STIX object that has the supplied STIX ID.
                The STIX object is loaded from its json file, parsed into
                a python STIX object and then returned

        """

        result = self.__retrieve_stix_object(stix_id)
        return result

    def query(self, query=None, version=None, _composite_filters=None):
        """Search and retrieve STIX objects based on the complete query.

        A "complete query" includes the filters from the query, the filters
        attached to this FileSystemSource, and any filters passed from a
        CompositeDataSource (i.e. _composite_filters).

        Args:
            query (list): list of filters to search on
            _composite_filters (FilterSet): collection of filters passed from
                the CompositeDataSource, not user supplied
            version (str): If present, it forces the parser to use the version
                provided. Otherwise, the library will make the best effort based
                on checking the "spec_version" property.

        Returns:
            (list): list of STIX objects that matches the supplied
                query. The STIX objects are loaded from their json files,
                parsed into a python STIX objects and then returned.

        """
        pass

    def all_versions(self, stix_id, version=None, _composite_filters=None):
        """Retrieve STIX object from file directory via STIX ID, all versions.

        Note: Since FileSystem sources/sinks don't handle multiple versions
        of a STIX object, this operation is unnecessary. Pass call to get().

        Args:
            stix_id (str): The STIX ID of the STIX objects to be retrieved.
            _composite_filters (FilterSet): collection of filters passed from
                the parent CompositeDataSource, not user supplied
            version (str): If present, it forces the parser to use the version
                provided. Otherwise, the library will make the best effort based
                on checking the "spec_version" property.

        Returns:
            (list): of STIX objects that has the supplied STIX ID.
                The STIX objects are loaded from their json files, parsed into
                a python STIX objects and then returned

        """
        pass



def get_embedded_match(source_id, import_type):
    """
        Assemble the typeql variable and match statement given the stix-id, and the increment
    Args:
        source_id (): stix-id to use
        i (): number of times this type of object has been used
    Returns:
        source_var, the typeql string of the variable
        match, the typeql match statement
    """
    source_type = get_source_type(source_id, import_type)
    source_var = '$' + source_type
    if source_type == 'relationship':
        source_type = 'stix-core-relationship'
    match = f' {source_var} isa {source_type}, has stix-id "{source_id}";\n'
    return source_var, match

def get_source_type(source_id, import_type):
    """
        Get the type of the source_id
    Args:
        source_id (): stix-id to use
    Returns:
        source_type, the type of the source_id
    """
    auth_factory = get_auth_factory_instance()
    auth = auth_factory.get_auth_for_import(import_type)
    tmp_source = source_id.split('--')[0]
    source = ""
    for model in auth["conv"]["sdo"]:
        if model["type"] == tmp_source:
            source = model["typeql"]
            return source
    for model in auth["conv"]["sro"]:
        if model["type"] == tmp_source:
            source = model["typeql"]
            return source
    for model in auth["conv"]["sco"]:
        if model["type"] == tmp_source:
            source = model["typeql"]
            return source
    for model in auth["conv"]["meta"]:
        if model["type"] == tmp_source:
            source = model["typeql"]
            return source
    return source



import functools
import importlib.util
import json
import logging
import os
import pathlib
import pkgutil
from typing import List, Dict
from stixorm.module import definitions
from stixorm.module.typedb_lib.factories.import_type_factory import ImportType
from stixorm.module.typedb_lib.factories.mappings_factory import get_mapping_factory_instance
from stixorm.module.typedb_lib.model.definitions import ModelDefinition, DefinitionName, ModelClassDefinition, \
    ImportTypeToDefinitionMapper


class DomainDefinition:

    def __init__(self,
                 domain_name: str,
                 path,
                 model_class: ModelClassDefinition):
        self.domain_name = domain_name
        self.definition_path = path

        base = self.__get_base()
        mappings = self.__get_mappings()
        sub_objects = self.__get_sub_objects()
        data = self.__get_data()
        # schema = self.__get_schema()
        # rules = self.__getrules()
        # vocab = self.__getvocab()
        # enums = self.__getenums()


        self.__model_definition = ModelDefinition(base=base,
                                                  mappings=mappings,
                                                  sub_objects=sub_objects,
                                                  data=data,
                                                  class_def=model_class)

    def is_json_file(self, filepath):
        _, file_extension = os.path.splitext(filepath)
        return file_extension.lower() == ".json"

    def __load_definition(self,
                          directory_path: str) -> dict:
        if not os.path.isdir(directory_path):
            return {}

        definitions = {}

        # Iterate over every file in the directory
        for filename in os.listdir(directory_path):
            filepath = os.path.join(directory_path, filename)
            if os.path.isfile(filepath) and self.is_json_file(filepath):
                # Do something with the file, for example, print its contents
                file_name, file_ext = os.path.splitext(filename)
                with open(filepath, 'r') as file:
                    definitions[file_name] = json.load(file)
        return definitions
    def __get_data(self) -> dict:
        # Define the directory path
        directory_path = str(self.definition_path.joinpath("data"))

        return self.__load_definition(directory_path)

    def __get_mappings(self) -> dict:

        # Define the directory path
        mappings_factory = get_mapping_factory_instance()

        return mappings_factory.get_mappings_from_definition(self.domain_name)

    def __get_sub_objects(self) -> dict:

        # Define the directory path
        directory_path = str(self.definition_path.joinpath("sub_objects"))

        return self.__load_definition(directory_path)

    def __get_base(self) -> dict:

        # Define the directory path
        directory_path = str(self.definition_path.joinpath("base"))

        return self.__load_definition(directory_path)


    def contains_mapping(self,
                         lookup: str) -> bool:
        return lookup in self.__model_definition.mappings.keys() and len(self.__model_definition.mappings[lookup]) > 0

   #def contains_class(self,
    #                     lookup: str) -> bool:
   #     return lookup in self.__model_definition.c.keys()

    def get_mapping(self,
                    lookup: str,
                    default = None):
        if default is None:
            return self.__model_definition.mappings.get(lookup)
        return self.__model_definition.mappings.get(lookup, default)

    def contains_data(self,
                      lookup: str) -> bool:
        return lookup in self.__model_definition.data.keys() and len(self.__model_definition.data[lookup]) > 0

    def contains_sub_objects(self,
                             lookup) -> bool:
        return lookup in self.__model_definition.sub_objects.keys() and len(self.__model_definition.sub_objects[lookup]) > 0

    def get_sub_objects(self,
                 lookup: str):
        return self.__model_definition.sub_objects[lookup]

    def get_data(self,
                 lookup: str):
        return self.__model_definition.data[lookup]

    def get_base(self,
                 lookup: str):
        return self.__model_definition.base[lookup]

    def does_property_contain_values(self,
                                   property: str) -> bool:
        model_def = self.__model_definition.dict()
        return property in model_def and len(model_def[property]) > 0

    def get_property_values(self,
                          property: str):
        model_def = self.__model_definition.dict()
        return model_def[property]

    def does_classes_contain_values(self,
                                   property: str) -> bool:
        model_def = self.__model_definition.class_def.dict()
        return property in model_def and len(model_def[property]) > 0

    def get_classes_property_values(self,
                          property: str):
        model_def = self.__model_definition.class_def.dict()
        return model_def[property]

class DefinitionFactory:

    def __init__(self):
        self.definition_dir = pathlib.Path(__file__).parents[2].joinpath('definitions')

        self.definitions: Dict[str, DomainDefinition] = {}

        package_name = definitions
        package = package_name.__name__
        for importer, subpackage_name, is_pkg in pkgutil.walk_packages(package_name.__path__, package + '.'):
            if is_pkg:
                try:
                    subpackage = importer.find_module(subpackage_name).load_module(subpackage_name)
                    name = subpackage.name
                    model_class = ModelClassDefinition(**subpackage.class_model)
                    self.definitions[name] = DomainDefinition(name, self.definition_dir.joinpath(name), model_class)
                except Exception as e:
                    logging.exception(e)


    def lookup_definition(self, domain_name: DefinitionName) -> DomainDefinition:
        if domain_name.value in self.definitions.keys():
            return self.definitions[domain_name.value]
        raise Exception("Definition for " + domain_name.value + " not found")


    def get_definitions_for_import(self,
                                   import_type: ImportType) -> List[DomainDefinition]:
        definitions = []
        for field_name, field_value in import_type.__dict__.items():
            definition_key = ImportTypeToDefinitionMapper().corresponding_definition_name(field_name)
            if field_value and definition_key is not None:
                definitions.append(self.definitions[definition_key.value])
        return definitions



@functools.lru_cache(maxsize=None)
def get_definition_factory_instance() -> DefinitionFactory:
    return DefinitionFactory()